#include <stdint.h>
#include "config.h"
#include "convert.h"
#include "dda.h"
#include "delay.h"
#include "dist.h"
#include "divide.h"
#include "log.h"
#include "motor.h"
#include "planner.h"
#include "sqrt.h"
#include "state.h"
#include "uart.h"

#define STATE_IDLE 0
#define STATE_START 1
#define STATE_SPEEDCHANGE 2
#define STATE_DDA 3
#define STATE_WAIT 4
#define STATE_FINISHED 5

#define MAX_ACCELERATION_XY ((uint32_t) (ACCELERATION_XY / (US_IN_S / PRECISION)))
#define MAX_ACCELERATION_Z ((uint32_t) (ACCELERATION_Z / (US_IN_S / PRECISION)))

#if DDA_SPEEDCHANGE_RATE == 1 || DDA_SPEEDCHANGE_RATE == 2 || DDA_SPEEDCHANGE_RATE == 4 || DDA_SPEEDCHANGE_RATE == 8 || DDA_SPEEDCHANGE_RATE == 16 || DDA_SPEEDCHANGE_RATE == 32 || DDA_SPEEDCHANGE_RATE == 64 || DDA_SPEEDCHANGE_RATE == 128 || DDA_SPEEDCHANGE_RATE == 256
#else
#error DDA_SPEEDCHANGE_RATE must be 1, 2, 4, 8, 16, 32, 64, 128 or 256
#endif

struct dda q[DDA_BUFFER];
uint8_t q_used = 0;

int32_t abs(int32_t x) {
    return x < 0 ? -x : x;
}

void dda_q_calculate_max_f_and_a(int32_t *fxyze, int32_t d, int32_t d_xy, int32_t d_z) {
    int32_t t_xy, t_z;

    t_xy = (d_xy << 8) / MAX_FEEDRATE_XY;
    t_z = (d_z << 8) / MAX_FEEDRATE_Z;
    if (t_xy < t_z)
        t_xy = t_z;
    if (t_xy < 1)
        t_xy = 1;
    q[q_used].f_max = ((d << 8) / t_xy);

    t_xy = d_xy / MAX_ACCELERATION_XY;
    t_z = d_z / MAX_ACCELERATION_Z;
    if (t_xy < t_z)
        t_xy = t_z;
    if (t_xy < 1)
        t_xy = 1;
    q[q_used].a = d / t_xy;
}

void dda_q(int32_t *fxyze) {
    static int32_t _xyze[5]; // in mm until acceleration plannig stage, then a unit vector
    static int32_t _xyze_old[5];
    int32_t d, d_xy;

    /* calculate movement coordinates in mm */
    FOREACH_XYZE(i) {
        _xyze_old[i] = _xyze[i];
        _xyze[i] = convert(fxyze[i], i + INV);
    }

    /* wait for buffer to have some space */
    while (q_used == DDA_BUFFER)
        dda_tick();

    /* calculate d and d_xy - the distances in XYZ and XY */
    d = dist(_xyze);
    d_xy = dist_xy(_xyze);
    if ((d>>8) == 0) {
        d = 0;
        FOREACH_XYZ(i)
            d += abs(_xyze[i]);
    }

    /* calculate the distance, maximum feedrate and acceleration
     * for this line segment */
    if (fxyze[X] == 0 && fxyze[Y] == 0 && fxyze[Z] == 0) {
        if (fxyze[E] == 0 || fxyze[E] == UNSEEN)
            return;
        d = abs(_xyze[E]);
        d_xy = 0;
        q[q_used].f_max = MAX_FEEDRATE_XY;
        q[q_used].a = MAX_ACCELERATION_XY;
    } else {
        if (fxyze[E] == UNSEEN) {
            fxyze[E] = convert(d, E);
            _xyze[E] = d;
        }
        dda_q_calculate_max_f_and_a(fxyze, d, d_xy, abs(_xyze[Z]));
    }

    FOREACH_XYZE(i) {
        _xyze[i] = (_xyze[i]<<8) / d;
        dda_tick();
    }

    q[q_used].d = d;
    if (q[q_used].f_max > fxyze[F])
        q[q_used].f_max = fxyze[F];
    FOREACH_XYZE(i) {
        q[q_used].delta[i] = fxyze[i];
        s.fxyze[i] += fxyze[i];
    }

#ifdef DEBUG
    uart_puts("dda_q ");
    uart_puti(q_used + 1);
    uart_puts("\t");
    FOREACH_XYZE(i) {
        uart_puts(" "); uart_puti(fxyze[i]);
    }
    uart_puts("\tf"); uart_puti(fxyze[F]);
    uart_puts("\td"); uart_putd(q[q_used].d);
    uart_puts("\tf"); uart_puti(q[q_used].f_max);
    uart_puts("\ta"); uart_puti(q[q_used].a*(US_IN_S/PRECISION));
    uart_puts("\n");
#endif

    planner(_xyze, _xyze_old); // plan acceleration
    q_used++;
}

void dda_tick(void) {
    static int8_t state = STATE_IDLE, shift;
    static int32_t error[5], max;
    static int32_t f = MIN_FEEDRATE * PRECISION;
    static int32_t f_max, df, a, x, x_d, advance, old_advance;
    static uint16_t dt;
    static int32_t steps_left = -1;
#ifdef DEBUG
    static int32_t count[5];
#endif
    int32_t e;
#if DDA_SPEEDCHANGE_RATE > 1
    static int8_t tick;
#endif

    switch(state++) {
    case STATE_IDLE:
        if (q_used == 0) {
            state = STATE_IDLE;

            if (old_advance > 0) {
                motor_steps(E, -old_advance);
                old_advance = 0;
            }
        }
        return;

    case STATE_START:
        delay_tick_reset();

        max = 0;
        steps_left = 0;
        x = 0;
        FOREACH_XYZE(i) {
            q[0].delta[i] = q[0].delta[i];
            motor_dir(i, q[0].delta[i] >= 0);
            if (q[0].delta[i] < 0)
                q[0].delta[i] = -q[0].delta[i];
            if (q[0].delta[i] != 0)
                motor_enable(i, 1);
            steps_left += q[0].delta[i];
            if (max < q[0].delta[i])
                max = q[0].delta[i];
        }

        f_max = q[0].f_max * PRECISION;
        a = q[0].a;
        shift = approx_log(f_max) - DDA_SHIFT;
        if (shift < 0) shift = 0;
        x_d = q[0].x_d >> shift;
        q[0].delta[T] = q[0].d >> shift;
        if (max < q[0].delta[T])
            max = q[0].delta[T];

        e = -(max >> 1);
        FOREACH_TXYZE(i)
            error[i] = e;
        return;

    case STATE_SPEEDCHANGE:
        /* calculate amount of time until the next T step (in us)
         *
         * dt = ((US_IN_MIN / T_steps_per_mm) << shift) / feedrate
         * where
         *      T_steps_per_mm = PRECISION
         *      feedrate = f / PRECISION
         * need to use >>8 to avoid overflow
         */
#ifdef DDA_USE_APPROX_DIVIDE
        dt = approx_divide((US_IN_MIN>>8) << shift, f>>8);
#else
        dt = ((US_IN_MIN>>8) << shift) / (f>>8);
#endif

        /* calculate amount of change to feedrate over that step */
        df = a * dt;

#ifdef DEBUG
        {
            static uint32_t old_f;
            if (f != old_f) {
                old_f = f;
                uart_puts("  x="); uart_putd(x<<shift);
                uart_puts(" f="); uart_putd(f);
                uart_puts(" dt="); uart_puti(dt);
                uart_puts(" df="); uart_putd(df);
                uart_puts("\n");
            }
        }
#endif
        return;

    case STATE_DDA:
        if (steps_left <= 0) {
            state = STATE_FINISHED;
            return;
        }

        state = STATE_DDA;
        FOREACH_TXYZE(a)
            error[a] += q[0].delta[a];

        FOREACH_XYZ(a)
            if (error[a] > 0) {
                error[a] -= max;
#ifdef DEBUG
                count[a]++;
#endif
                motor_step(a);
                steps_left--;
            }
        motor_step_done();

        e = 0;
        if (error[E] > 0) {
            error[E] -= max;
#ifdef DEBUG
            count[E]++;
#endif
            e++;
            steps_left--;
        }
        if (q[0].delta[E]) {
            /*
             * advance = STEPS_PER_CUBIC_MM_E * K * v^2
             * where
             *     v = f * area
             *     f = (f * 256 / PRECISION / 60)
             *     area = s.extrusion_area / PRECISION
             */

            /*
             * advance = (s.extruder_advance * rate^2) >> 16
             * rate = (f>>12 * s.extrusion_area) >> 16
             *      = (f*60*PRECISION>>12 * area*PRECISION) >> 16
             *      = (f*area * 60*PRECISION*PRECISION) >> 28
             * rate^2 = ((f*area)^2 * 60^2*PRECISION^4) >> 56
             *
             * therefore...
             * s.extruder_advance
             *     = ((STEPS_PER_CUBIC_MM * K) << 72)
             *       / (60^2 * PRECISION^4)
             *     = STEPS_PER_CUBIC_MM * K * PRECISION_ADVANCE
             */
            int32_t rate = ((f>>12) * s.extrusion_area) >> 16;
            advance = (s.extruder_advance * rate * rate) >> 16;

#if DEBUG
            if (advance != old_advance) {
                uart_puts("    f="); uart_putd(f);
                uart_puts(" area="); uart_putd(s.extrusion_area);
                uart_puts(" rate="); uart_puti(rate);
                uart_puts(" advance="); uart_puti(advance);
                uart_puts("\n");
            }
#endif
        } else {
            advance = 0;
        }
        motor_steps(E, e + advance - old_advance);
        old_advance = advance;

        if (error[T] > 0) {
            x++;
            error[T] -= max;
            state = STATE_WAIT;
            delay_tick_in(dt);
        }
        return;

    case STATE_WAIT:
        if (!delay_ticked()) {
            state = STATE_WAIT;
            return;
        }

#if DDA_SPEEDCHANGE_RATE == 1
        state = STATE_SPEEDCHANGE;
#else
        if ((++tick & (DDA_SPEEDCHANGE_RATE - 1)) == 0)
            state = STATE_SPEEDCHANGE;
        else
            state = STATE_DDA;
#endif

        if (x > x_d)
            f -= df;
        else
            f += df;
        if (f < MIN_FEEDRATE * PRECISION)
            f = MIN_FEEDRATE * PRECISION;
        if (f > f_max)
            f = f_max;
        return;

    case STATE_FINISHED:
#ifdef DEBUG
        uart_puts("dda_done \t");
        FOREACH_XYZE(i) {
            uart_puts(" ");
            uart_puti(count[i]);
            count[i] = 0;
        }
        uart_puts("\n");
#endif

        if (AUTO_DISABLE_X) motor_enable(X, 0);
        if (AUTO_DISABLE_Y) motor_enable(Y, 0);
        if (AUTO_DISABLE_Z) motor_enable(Z, 0);
        if (AUTO_DISABLE_E) motor_enable(E, 0);

        f = q[0].f_exit * PRECISION;

        for (uint8_t i = 1; i < q_used; i++)
            q[i-1] = q[i];
        q_used--;
        state = STATE_IDLE;
        return;
    }
}

void dda_drain(void) {
    while (q_used > 0)
        dda_tick();
}
