#include <stdint.h>
#include "config.h"
#include "dda.h"
#include "sqrt.h"
#include "state.h"
#include "uart.h"

/* calculate the distance travelled when accelerating from feedrate 1 to feedrate 2*/
int32_t planner_accel_distance(int32_t f1, int32_t f2, int32_t a) {
//    int32_t f_avg = (f1 + f2) / 2;
//    int32_t t = PRECISION * (f2 - f1) / a;
//    return f_avg * t / 60;
    return (f2 * f2 - f1 * f1) / (a * 60 * 2 / PRECISION);
}

/* calculate the speed after accelerating for a distance */
uint32_t planner_feedrate_accel(int32_t f1, int32_t d, int32_t a) {
    if (d > 10 * PRECISION)
        return MAX_FEEDRATE_XY;
    return ((uint32_t)sqrt32((2 * 60 * a / PRECISION) * d + f1 * f1));
}

/* calculate the maximum corner feedrate to keep the jolt small */
uint32_t planner_feedrate_jolt(int32_t *_xyze1, int32_t *_xyze2) {
    // if we are starting (or stopping) extruding, slow to MIN_FEEDRATE
    if ((_xyze1[E] == 0) ^ (_xyze2[E] == 0))
        return MIN_FEEDRATE;

    // cos(a) = (x1 * x2 + y1 * y2) / (d1 * d2)
    // sin(a) = (x1 * y2 - x2 * y1) / (d1 * d2)
    // ... but d1 and d2 are 1

    // jolt = sqrt(jolt_a^2 + jolt_b^2 + jolt_z^2)
    //   jolt_a = 1 - (x1 * x2 + y1 * y2)
    //   jolt_b = (x1 * y2 - x2 * y1)
    //   jolt_z = z1 - z2

    // jolt has range 0..2 (straight on to straight back)
    // so feedrate we can accept is JOLT_FEEDRATE * 2 / jolt
    uint32_t jolt_a = (1<<8) - ((_xyze1[X]*_xyze2[X] + _xyze1[Y]*_xyze2[Y])>>8);
    uint32_t jolt_b = (_xyze1[X]*_xyze2[Y] + _xyze1[Y]*_xyze2[X])>>8;
    uint32_t jolt_z = _xyze1[Z] - _xyze2[Z];
    uint16_t jolt = sqrt32(jolt_a*jolt_a + jolt_b*jolt_b + jolt_z*jolt_z);
    if (jolt == 0)
        jolt = 1;
    return ((JOLT_FEEDRATE * 2)<<8) / jolt;
}

void planner(int32_t *_xyze1, int32_t *_xyze2) {
    int32_t a, f_entry, f_exit;

    /* plan acceleration and deceleration points
     *
     * calculate forwards the maximum corner speeds so that the jerk at
     * the corners is acceptable and we do not need to accelerate by
     * more than a.
     *
     * then calculating backwards limit those speeds so that we do not
     * need to deccelerate by more than a.  Assume that we finish the
     * last segement at the minimum feedrate (MIN_FEEDRATE).
     *
     * Finally, calculate the position to decelerate (x_d) for each
     * segment.
     *
     */

    /* if we have a new corner... set f_exit_max of the penultimate
     * segment to the maximum corner feedrate */
    if (q_used > 0) {
        a = planner_feedrate_jolt(_xyze1, _xyze2);
        if (a > q[q_used].f_max) a = q[q_used].f_max;
        if (a > q[q_used-1].f_max) a = q[q_used-1].f_max;
        q[q_used-1].f_exit_max = a;
    }

    /* forwards */
    f_exit = q[0].f_exit;
    for (uint8_t i = 1; i <= q_used; i++) {
        a = q[i].a * (US_IN_S/PRECISION);
        f_entry = f_exit;

        /* limit f_exit to maximum rate possible given acceleration */
        f_exit = planner_feedrate_accel(f_entry, q[i].d, a);
        if (f_exit > q[i].f_exit_max) f_exit = q[i].f_exit_max;

        /* store */
        q[i].f_exit = f_exit;
    }

    /* backwards */
    q[q_used].f_exit = f_entry = MIN_FEEDRATE;
    for (uint8_t i = q_used; i > 0; i--) {
        a = q[i].a * (US_IN_S/PRECISION);
        f_exit = f_entry;

        /* store */
        if (f_exit < q[i].f_exit)
            q[i].f_exit = f_exit;

        /* limit f_entry to maximum rate possible given decceleration */
        f_entry = planner_feedrate_accel(q[i].f_exit, q[i].d, a);
    }

    /* forwards: update decceleration points */
    f_exit = q[0].f_exit; // will be MIN_FEEDRATE if q_used == 0
    for (uint8_t i = q_used > 0 ? 1 : 0; i <= q_used; i++) {
        a = q[i].a * (US_IN_S/PRECISION);
        f_entry = f_exit;
        f_exit = q[i].f_exit;

        /* calculate (and save) x_d for this segment */
        int32_t x_a = planner_accel_distance(f_entry, q[i].f_max, a);
        int32_t x_d = q[i].d - planner_accel_distance(f_exit, q[i].f_max, a);
        if (x_d >= x_a)
            q[i].x_d = x_d;
        else
            q[i].x_d = (x_a + x_d) / 2;
    }

#ifdef DEBUG
    uart_puts("plan:\n");
    for (uint8_t i = 0; i <= q_used; i++) {
        uart_puts("  "); uart_puti(i);
        uart_puts(" f_max="); uart_puti(q[i].f_max);
        uart_puts(" f_exit="); uart_puti(q[i].f_exit);
        uart_puts(" f_exit_max="); uart_puti(q[i].f_exit_max);
        uart_puts(" d="); uart_putd(q[i].d);
        uart_puts(" x_d="); uart_putd(q[i].x_d);
        uart_puts("\n");
    }
#endif
}
